/*
 * Copyright 2016, Simula Research Laboratory
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#pragma once

#define POPSIFT_VERSION_MAJOR 0
#define POPSIFT_VERSION_MINOR 10
#define POPSIFT_VERSION_PATCH 1

#define POPSIFT_VERSION_STRING "0.10.1"
